/*
 *      ipvsadmcgi.c - IP Virtual Server ADMinistration CGI-bin program v1.0
 *
 *      Copyright (c) 1998 Wensong Zhang
 *      All rights reserved.
 *
 *      Author: Wensong Zhang <wensong@iinchina.net>
 *
 *
 *      This program is free software; you can redistribute it and/or modify
 *      it under the terms of the GNU General Public License as published by
 *      the Free Software Foundation; either version 2 of the License, or
 *      (at your option) any later version.
 *
 *      This program is distributed in the hope that it will be useful,
 *      but WITHOUT ANY WARRANTY; without even the implied warranty of
 *      MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *      GNU General Public License for more details.
 *
 *      You should have received a copy of the GNU General Public License
 *      along with this program; if not, write to the Free Software
 *      Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 *
 */

#include <stdio.h>
#include <errno.h>
#include <limits.h>
#include <string.h>
#include <stdlib.h>
#include <unistd.h>
#include <getopt.h>
#include <netdb.h>
#include <netinet/in.h>
#include <linux/ip.h>
#include <linux/tcp.h>
#include <linux/udp.h>
#include <linux/icmp.h>
#include <linux/if.h>
#include <linux/timer.h>
#include <linux/ip_fw.h>
#include <sys/param.h>

#define IP_PORTFW_NONE	0
#define IP_PORTFW_LIST	10000
#define DEFAULT_WEIGHT      1
#define IPPROTO_NONE	65535

long string_to_number (char *str, int min, int max);
int do_setsockopt (int cmd, struct ip_portfw_edits *data, int length);
void exit_error (int status, char *msg);
void exit_display_help (void);
void list_forwarding (void);
void parse_parameters (struct ip_portfw_edits *pfw, int *command);
void print_html_header (void);
void print_html_footer (void);
void getword (char *word, char *line, char stop);
void escape_shell_cmd (char *cmd);
char x2c (char *what);
void unescape_url (char *url);
void plustospace (char *str);


void
main (int argc, char *argv[])
{
  int c;
  int command = IP_PORTFW_NONE;
  struct ip_portfw_edits pfw;

  pfw.protocol = IPPROTO_NONE;
  pfw.raddr = 0;
  pfw.rport = 0;
  pfw.laddr = 0;
  pfw.lport = 0;
  pfw.weight = 0;


  parse_parameters (&pfw, &command);



  if (command == IP_PORTFW_NONE)
    {
      print_html_header ();
      list_forwarding ();
      print_html_footer ();
    }
  else
    {
      printf ("Content-type: text/html\n\n\n");
      printf ("<html>\n");
      printf ("<head><title>Virtual Server Administration</title>\n");
      printf ("<meta http-equiv=Refresh content=\"5; url=/cgi-bin/ippfvsadm.cgi\">\n");
      printf ("<body>\n");

      if (do_setsockopt (command, &pfw, sizeof (pfw)) == 0)
	printf ("Succeeded...\n");
      else
	printf ("<p>Failed...<p>Please go back and check parameters you entered!\n");

      printf ("</body></html>");
    }

}


long
string_to_number (char *str, int min, int max)
{
  char *end;
  long number;

  number = strtol (str, &end, 10);
  if (*end == '\0' && end != str)
    {
      if (min <= number && number <= max)
	return number;
      else
	return -1;
    }
  else
    return -1;
}


int
get_address (char *str, __u32 * raddr)
{
  char buf[23];			/* xxx.xxx.xxx.xxx\0 */
  char *p, *q;
  int onebyte, i;

  strcpy (buf, str);

  p = buf;
  *raddr = 0;
  for (i = 0; i < 3; i++)
    {
      if ((q = strchr (p, '.')) == NULL)
	return -1;
      else
	{
	  *q = '\0';
	  if ((onebyte = string_to_number (p, 0, 255)) == -1)
	    return -1;
	  else
	    *raddr = (*raddr << 8) + onebyte;
	}
      p = q + 1;
    }

  /* we've checked 3 bytes, now we check the last one */
  if ((onebyte = string_to_number (p, 0, 255)) == -1)
    return -1;
  else
    *raddr = (*raddr << 8) + onebyte;

  return 0;
}


int
do_setsockopt (int cmd, struct ip_portfw_edits *data, int length)
{
  static int sockfd = -1;
  int ret;

  if (sockfd == -1)
    {
      if ((sockfd = socket (AF_INET, SOCK_RAW, IPPROTO_RAW)) == -1)
	{
	  perror ("ipfwadm: socket creation failed");
	  exit (1);
	}
    }

  ret = setsockopt (sockfd, IPPROTO_IP, cmd, (char *) data, length);
  if (ret)
    perror ("ipfwadm: setsockopt failed");

  return ret;
}


void
list_forwarding (void)
{
  char buffer[256];

  FILE *handle;
  handle = fopen ("/proc/net/ip_portfw", "r");
  if (!handle)
    {
      printf ("Could not open /proc/net/ip_portfw\n");
      printf ("Are you sure you have Port Forwarding & Virtual Server support installed?\n");
      exit (1);
    }

  while (!feof (handle))
    if (fgets (buffer, 256, handle))
      printf ("%s", buffer);
  fclose (handle);

}


void
parse_parameters (struct ip_portfw_edits *pfw, int *command)
{
  char wd[64];
  char ArgOfQuery[1024];

  /* Parse the environment variables */
  strncpy (ArgOfQuery, getenv ("QUERY_STRING"), sizeof (char) * 1024);

  while (ArgOfQuery[0] != '\0')
    {
      getword (wd, ArgOfQuery, '=');
      if (!strcmp (wd, "laddr"))
	{
	  getword (wd, ArgOfQuery, '&');
	  plustospace (wd);
	  unescape_url (wd);
	  get_address (wd, &pfw->laddr);
	}
      else if (!strcmp (wd, "lport"))
	{
	  getword (wd, ArgOfQuery, '&');
	  plustospace (wd);
	  unescape_url (wd);
	  pfw->lport = string_to_number (wd, IP_PORTFW_PORT_MIN, IP_PORTFW_PORT_MAX);
	}
      else if (!strcmp (wd, "raddr"))
	{
	  getword (wd, ArgOfQuery, '&');
	  plustospace (wd);
	  unescape_url (wd);
	  get_address (wd, &pfw->raddr);
	}
      else if (!strcmp (wd, "rport"))
	{
	  getword (wd, ArgOfQuery, '&');
	  plustospace (wd);
	  unescape_url (wd);
	  pfw->rport = string_to_number (wd, IP_PORTFW_PORT_MIN, IP_PORTFW_PORT_MAX);
	}
      else if (!strcmp (wd, "weight"))
	{
	  getword (wd, ArgOfQuery, '&');
	  plustospace (wd);
	  unescape_url (wd);
	  pfw->weight = string_to_number (wd, SERVER_WEIGHT_MIN, SERVER_WEIGHT_MAX);
	}
      else if (!strcmp (wd, "add"))
	{
	  *command = IP_PORTFW_ADD;
	}
      else if (!strcmp (wd, "delete"))
	{
	  *command = IP_PORTFW_DEL;
	}
      else if (!strcmp (wd, "proto"))
	{
	  getword (wd, ArgOfQuery, '&');
	  if (wd[0] == 'Y')
	    pfw->protocol = IPPROTO_TCP;
	  else
	    pfw->protocol = IPPROTO_UDP;

	}
    }
}


void
print_html_header (void)
{
  printf ("Content-type: text/html\n\n\n");
  printf ("<HTML>\n"
	  "<HEAD>\n"
	  "<META HTTP-EQUIV=\"Content-Type\" CONTENT=\"text/html; charset=iso-8859-1\">\n"
	  "<META NAME=\"Author\" CONTENT=\"Wensong Zhang\">\n"
	  "<TITLE>Virtual Server Administration</TITLE>\n"
	  "</HEAD>\n"
	  "<BODY>\n"
	  "<UL>\n"
	  "<LI>\n"
	  "<FONT SIZE=+1>Current Virtual Server Rules:</FONT></LI>\n"
	  "</UL>\n"
	  "<HR WIDTH=\"100%\">\n"
	  "<PRE>\n");
}


void
print_html_footer (void)
{
  printf ("</PRE>\n"
	  "<HR WIDTH=\"100%\">\n"
	  "<UL>\n"
	  "<LI>\n"
	  "Add/Delete A Rule:</LI>\n"
	  "</UL>\n"
	  "<FORM method=get action=\"/cgi-bin/ippfvsadm.cgi\">\n"
	  "<CENTER><TABLE COLS=6 WIDTH=\"75%\" >\n"
	  "<TR>\n"
	  "<TD>Protocol</TD>\n"
	  "<TD>Virtual Server</TD>\n"
	  "<TD>Port</TD>\n"
	  "<TD>Real Server</TD>\n"
	  "<TD>Port</TD>\n"
	  "<TD>Weight</TD>\n"
	  "</TR>\n"
	  "<TR>\n"
	  "<TD><INPUT type=radio name=proto value=Y checked>TCP<INPUT type=radio name=proto value=N>UDP</TD>\n"
	  "<TD><INPUT type=text name=laddr size=15 maxlength=15></TD>\n"
	  "<TD><INPUT type=text name=lport size=8 maxlength=8></TD>\n"
	  "<TD><INPUT type=text name=raddr size=15 maxlength=15></TD>\n"
	  "<TD><INPUT type=text name=rport size=8 maxlength=8></TD>\n"
	  "<TD><INPUT type=text name=weight size=10 maxlength=8></TD>\n"
	  "</TR>\n"
	  "</TABLE></CENTER>\n"
	  "<CENTER>&nbsp;</CENTER>\n"
	  "<CENTER><TABLE COLS=3 WIDTH=\"32%\" >\n"
	  "<TR>\n"
	  "<TD><INPUT type=submit name=add value=\"   Add   \"></TD>\n"
     "<TD><INPUT type=submit name=delete size=10 value=\" Delete \"></TD>\n"
	  "<TD><INPUT type=reset size=20 value=\"  Reset \"></TD>\n"
	  "</TR>\n"
	  "</TABLE></CENTER>\n"
	  "<CENTER></FORM></CENTER>\n"
	  "</BODY>\n"
	  "</HTML>\n");
}


void
getword (char *word, char *line, char stop)
{
  register int x = 0, y;

  for (x = 0; ((line[x]) && (line[x] != stop)); x++)
    word[x] = line[x];

  word[x] = '\0';
  if (line[x])
    ++x;
  y = 0;

  while (line[y++] = line[x++]);
}


char
x2c (char *what)
{
  register char digit;

  digit = (what[0] >= 'A' ? ((what[0] & 0xdf) - 'A') + 10 : (what[0] - '0'));
  digit *= 16;
  digit += (what[1] >= 'A' ? ((what[1] & 0xdf) - 'A') + 10 : (what[1] - '0'));
  return (digit);
}


void
unescape_url (char *url)
{
  register int x, y;

  for (x = 0, y = 0; url[y]; ++x, ++y)
    {
      if ((url[x] = url[y]) == '%')
	{
	  url[x] = x2c (&url[y + 1]);
	  y += 2;
	}
    }
  url[x] = '\0';
}


void
plustospace (char *str)
{
  register int x;

  for (x = 0; str[x]; x++)
    if (str[x] == '+')
      str[x] = ' ';
}
